def test(self):
    import numpy as np
    from .unsatfit import Fit
    f = Fit()
    # Test data from UNSODA 3393
    f.swrc = (np.array([10, 28, 74, 160, 288, 640, 1250, 2950, 6300, 10600, 15800]), np.array(
        [0.36, 0.35, 0.34, 0.33, 0.32, 0.3, 0.28, 0.26, 0.24, 0.22, 0.2]))
    f.unsat = (np.array([10, 28, 74, 160, 288, 640, 1250, 2950, 6300, 10600]), np.array(
        [0.384, 0.0988, 0.0293, 0.0137, 0.00704, 0.00315, 0.00085, 0.000206, 0.000101, 0.00006]) / 60 / 60 / 24)
    f.ini = (max(f.unsat[1]), 1.5)  # Initial values of Ks and p
    f.set_model('Brooks and Corey', const=[f.get_wrf_bc(), 'q=1', 'r=2'])
    f.test_confirm(829)
    f.set_model('van Genuchten', const=[f.get_wrf_vg(), 'r=2'])
    f.modified_model(2)
    f.test_confirm(922)
    f.set_model('Kosugi', const=[f.get_wrf_ln(), 'q=1', 'r=2'])
    f.test_confirm(954)
    f.set_model('dual-VG-CH', const=[f.get_wrf_vg2ch(), 'r=2'])
    f.test_confirm(893)
    f.set_model('KO1BC2-CH', const=[f.get_wrf_kobcch(), 'q=1', 'r=2'])
    f.test_confirm(966)
    f.set_model('KO1BC2', const=[f.get_wrf_kobc(), 'q=1', 'r=2'])
    f.modified_model(2)
    f.test_confirm(856)


def test_confirm(self, expect):
    self.optimize()
    result = int((self.r2_ht + self.r2_ln_hk) * 500)
    assert abs(expect - result) < 2, 'Test failed for {0}. Expected: {1} Actual: {2}\nResult: {3}'.format(
        self.model_name, expect, result, self.message)



# van Genuchten model
import numpy as np


def init_model_vg(self):
    self.model['van Genuchten'] = self.model['VG'] = self.model['vg'] = {
        'function': (self.vg, self.vg_k),
        'bound': self.bound_vg,
        'get_init': self.get_init_vg,
        'get_wrf': self.get_wrf_vg,
        'param': ['qs', 'qr', 'a', 'm', 'Ks', 'p', 'q', 'r'],
        'k-only': [4, 5, 7]
    }
    self.model['MVG'] = self.model['Modified VG'] = self.model['mvg'] = {
        'function': (self.mvg, self.mvg_k),
        'bound': self.bound_mvg,
        'param': ['qs', 'qr', 'a', 'm', 'hs', 'Ks', 'p', 'q', 'r'],
        'k-only': [5, 6, 8]
    }


def bound_vg(self):
    return [self.b_qs, self.b_qr, self.b_a, self.b_m,
            self.b_ks, self.b_p, self.b_q, self.b_r]


def vg(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.vg_se(p[2:5], x) * (p[0] - p[1]) + p[1]


def vg_se(self, p, x):
    a, m, q = p
    n = q / (1 - m)
    return (1 + (a * x)**n)**(-m)


def vg_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, a, m, ks, p, q, r = par
    s = self.vg_se([a, m, q], x)
    k = ks * s**p * (1 - (1 - s**(1 / m))**m)**r
    return k


def get_init_vg(self):  # alpha and m
    from .unsatfit import Fit
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.set_model('vg', const=[[1, 1], [2, 0], [7, 1]])
    f.swrc = (x, y)
    hb, l = f.get_init_bc()
    n = l + 1
    f.ini = (1 / hb, 1 - 1 / n)
    f.optimize()
    return f.fitted


def get_wrf_vg(self):
    from .unsatfit import Fit
    f = Fit()
    f.swrc = self.swrc
    f.debug = self.debug
    a, m = f.get_init_vg()
    f.set_model('vg', const=['q=1'])
    qs = max(f.swrc[1])
    f.ini = (qs, 0, a, m)
    f.b_a = (0, a * 100)
    f.optimize()
    if f.success:
        return (*f.fitted, 1)
    f.b_qs = (qs * 0.95, qs * 1.5)
    f.b_qr = (0, min(f.swrc[1]))
    f.optimize()
    if f.success:
        return (*f.fitted, 1)
    return (*f.ini, 1)

# Modified van Genuchten model (Vogel, 2000)


def bound_mvg(self):
    return [self.b_qs, self.b_qr, self.b_a, self.b_m,
            self.b_hs, self.b_ks, self.b_p, self.b_q, self.b_r]


def mvg(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.mvg_se(p[2:6], x) * (p[0] - p[1]) + p[1]


def mvg_se(self, p, x):
    a, m, hs, q = p
    n = q / (1 - m)
    sm = (1 + (a * hs)**n)**(m)
    return np.where(x < hs, 1, sm * (1 + (a * x)**n)**(-m))


def mvg_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, a, m, hs, ks, p, q, r = par
    s = self.vg_se([a, m, q], x)
    shs = self.vg_se([a, m, q], hs)
    ah = (1 - (1 - s**(1 / m))**m) / (1 - (1 - shs**(1 / m))**m)
    return np.where(x < hs, ks, ks * s**p * ah**r)



def init_bound(self):
    """Set initial bound of parameters

    Each bound is set as tuple of (minimum, maximum)
    """
    import numpy as np
    self.b_qs = self.b_qr = (0, np.inf)
    self.b_qa = self.b_w1 = self.b_m = (0, 1)
    self.b_a = self.b_a1 = self.b_a2 = self.b_hm = self.b_hm1 = self.b_hm2 = self.b_sigma = self.b_he = (
        0, np.inf)
    self.b_hb = self.b_hb2 = self.b_hc = self.b_hs = self.b_lambda = self.b_lambda1 = self.b_lambda2 = (
        0, np.inf)
    self.b_fxa = self.b_fxm = self.b_fxn = (0, np.inf)
    self.b_ks = self.b_p = self.b_q = self.b_r = (0, np.inf)



class Fit:
    """Fit water retention and unsaturated hydraulic conductivity functions

    unsatfit is a Python library for optimizing parameters of functions of
    soil hydraulic properties (water retention function and unsaturated
    hydraulic conductivity function).

    See document at https://sekika.github.io/unsatfit/

    PyPI: https://pypi.org/project/unsatfit/
    Source: https://github.com/sekika/unsatfit
    Author: Katsutoshi Seki
    License: MIT License

    Reference

    Seki, K., Toride, N., & Th. van Genuchten, M. (2022) Closed-form hydraulic
        conductivity equations for multimodal unsaturated soil hydraulic properties.
        Vadose Zone J. 21, e20168. https://doi.org/10.1002/vzj2.20168
    """

    from ._init_model import init_model, set_model, get_init_not_defined, get_wrf_not_defined
    from ._init_bound import init_bound
    from ._init_lsq import init_lsq
    from ._init_fig import init_fig

    from ._model_bc import init_model_bc, bound_bc, bc, bc_se, bc_k, get_init_bc, get_wrf_bc
    from ._model_vg import init_model_vg, bound_vg, vg, vg_se, vg_k, get_init_vg, get_wrf_vg, bound_mvg, mvg, mvg_se, mvg_k
    from ._model_ko import init_model_ln, bound_ln, ln, ln_se, ln_k, get_init_ln, get_wrf_ln
    from ._model_fx import init_model_fx, bound_fx, fx, fx_se, get_init_fx
    from ._model_dual_bc import init_model_bc2, bound_bc2f, bc2f, bc2f_se, bc2f_k, bound_bc2, bc2, bc2_se, bc2_k, get_init_bc2, get_wrf_bc2, bc2ca_k
    from ._model_dual_vg import init_model_vg2, bound_vg2, vg2, vg2_se, vg2_k, get_init_vg2, get_wrf_vg2, bound_vg2ch, vg2ch, vg2ch_se, vg2ch_k, get_init_vg2ch, get_wrf_vg2ch, vg2chca_k
    from ._model_dual_ko import init_model_ln2, bound_ln2, ln2, ln2_se, ln2_k, bound_ln2ch, ln2ch, ln2ch_se, ln2ch_k
    from ._model_vgbc import init_model_vgbc, bound_vgbc, vgbc, vgbc_se, vgbc_k, bound_vgbcp2, vgbcp2_k, bound_vgbcch, vgbcch, vgbcch_se, vgbcch_k, get_init_vgbcch, bound_vgbcchp2, vgbcchp2, vgbcchp2_k
    from ._model_kobc import init_model_kobc, bound_kobc, kobc, kobc_se, kobc_k, get_wrf_kobc, bound_kobcp2, kobcp2_k, bound_kobcch, kobcch, kobcch_se, kobcch_k, get_init_kobcch, get_wrf_kobcch, bound_kobcchp2, kobcchp2_k, kobcchca_k
    from ._model_fs import init_model_fs, bound_vgfs, vgfs, vgfs_se, vgfs_k
    from ._model_pe import init_model_pe, bound_pk, pk, pk_se, pk_k, get_init_pk, get_wrf_pk
    from ._model_modified import modified_model, modified_ht, modified_hk

    from ._optimize import optimize, multi_ini, format, residual_ht, residual_ln_hk, residual_log10_hk, f_r2_ht, f_r2_ln_hk, p_ht, total_cost
    from ._figure import set_scale, add_curve, clear_curves, h_0to1, plot
    from ._contour import contour, label

    from ._test import test, test_confirm

    def __init__(self):
        self.debug = False
        self.swrc = self.unsat = []  # Set empty data
        self.init_model()  # Define soil hydraulic models
        self.init_bound()  # Boundary conditions
        self.init_lsq()   # Parameters for least square optimization
        self.init_fig()   # Parameters for figure

    def version(self):
        """get version"""
        import configparser
        import os
        inifile = configparser.ConfigParser()
        here = os.path.abspath(os.path.dirname(__file__))
        inifile.read(os.path.join(here, 'data/system.ini'))
        return inifile.get('system', 'version')

    def linear_regress(self, x, y):
        """Linear regression y = ax"""
        import numpy as np
        return np.dot(x, y) / (x**2).sum()



# Brooks and Corey model
import numpy as np


def init_model_bc(self):
    self.model['Brooks and Corey'] = self.model['BC'] = self.model['bc'] = {
        # Water retention and hydraulic conductivity functions
        'function': (self.bc, self.bc_k),
        # Bound of paramters
        'bound': self.bound_bc,
        # Name of parameters
        'param': ['qs', 'qr', 'hb', 'l', 'Ks', 'p', 'q', 'r'],
        # Function to get initial WRF parameters except qs, qr
        'get_init': self.get_init_bc,
        # Function to get initial WRF parameters
        'get_wrf': self.get_wrf_bc,
        # Index of parameters (starting from 0) used only for K function
        'k-only': [4, 5, 6, 7]
    }


def bound_bc(self):
    return [self.b_qs, self.b_qr, self.b_hb, self.b_lambda,
            self.b_ks, self.b_p, self.b_q, self.b_r]


def bc(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.bc_se(p, x) * (p[0] - p[1]) + p[1]


def bc_se(self, p, x):
    # Ignore runtime warning, because divide by zero is warned when x=0
    import warnings
    warnings.simplefilter('ignore', category=RuntimeWarning)
    return np.where(x < p[2], 1, (x / p[2]) ** (-p[3]))


def bc_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, hb, l, ks, p, q, r = par
    s = self.bc_se(par[:4], x)
    k = ks * s**p * (np.where(x < hb, 1, (x / hb) ** (-l - q)))**r
    return k


def get_init_bc(self):  # hb and lambda
    import math
    from .unsatfit import Fit
    swrc = list(zip(*self.swrc))
    swrc_sort = sorted(swrc, key=lambda x: x[0])
    swrc = list(zip(*swrc_sort))
    x, t = swrc
    y = t / max(t)
    hbi = sum(y > 0.95 + min(y) * 0.05)
    hli = sum(y > 0.15 + min(y) * 0.85)
    if hbi == hli:
        hli = hli + 1
    if hbi > hli:
        hbi = hbi - 1
        hli = hbi + 1
    hb = x[hbi]
    if hb == 0:
        hb = x[2] / 10
    l = -math.log(y[hli] / y[hbi]) / math.log(x[hli] / hb)
    if l < 0.1:
        l = 0.1
    if l > 10:
        l = 10
    f = Fit()
    f.set_model('bc', const=[[1, 1], [2, 0]])
    f.swrc = (x, y)
    f.ini = (hb, l)
    f.optimize()
    hb, l = f.fitted
    r2 = f.r2_ht
    if r2 < 0.6:
        f.set_model('bc', const=[[2, 0]])
        f.ini = (1, *f.ini)
        f.optimize()
        if f.r2_ht > r2:
            hb, l = f.fitted[1:]
    return hb, l


def get_wrf_bc(self):
    from .unsatfit import Fit
    f = Fit()
    f.swrc = self.swrc
    f.debug = self.debug
    hb, l = f.get_init_bc()
    f.set_model('bc', const=[])
    f.ini = (max(f.swrc[1]), 0, hb, l)
    f.optimize()
    if f.success:
        return f.fitted
    return f.ini



def init_fig(self):
    """Set initial parameters for drawing figure"""
    self.show_fig = False  # Show figure by matplotlib.pyplot.show
    # Save figure as self.filename where output format is deduced from the extension of the filename
    self.save_fig = False
    self.fig_h_0to1 = False  # Change data of h=0 to h=1 for logarithmic plot
    self.data_only = False  # Plot data only
    self.have_new_plot = True

    # Size of the figure in inch
    self.fig_width = 4.3
    self.fig_height = 3.2
    self.fig_height_double = 4.7
    self.top_margin = 0.05
    self.bottom_margin = 0.12
    self.left_margin = 0.17  # Space for label is needed
    self.right_margin = 0.05
    self.hspace = 0.07

    # Axis
    self.log_x = True  # Logarithmic x axis
    self.min_x_log = 1  # Minimum of x axis in logrithmic scale
    # self.max_x = 10**4
    # self.max_y1 = 0.5
    # self.min_y2 = 10**(-12)
    self.max_y2_log = 5  # Maximum of K axis / maximum of K value

    # Marker
    self.marker = 'o'  # Select from https://matplotlib.org/stable/api/markers_api.html
    # See https://matplotlib.org/stable/gallery/color/named_colors.html
    self.color_marker = 'black'

    # Curves
    # Order of colors and styles
    self.color = ('red', 'blue', 'green', 'magenta', 'cyan', 'black')
    # See https://matplotlib.org/stable/gallery/lines_bars_and_markers/linestyles.html
    self.style = ['dashed', 'dashdot', 'solid', 'dotted']
    self.curve_smooth = 200  # Numbers of calculation points in a curve
    self.line_num = 0
    self.curves_ht = []
    self.curves_hk = []

    # Label of axis
    self.label_head = 'Matric head'
    self.label_theta = 'Volumetric water content'
    self.label_k = 'Hydraulic conductivity'

    # Legend
    self.legend = True
    self.show_r2 = True
    self.legend_fontsize = 10
    self.legend_loc = 'center right'
    self.legend_facecolor = 'white'
    self.legend_opacity = 1  # 0 = transparent, 1 = opaque
    self.data_legend = 'Measured'
    self.line_legend = 'Fitted'

    # Contour
    self.contour_level = 8
    self.contour_color = None  # specified by cmap. specify a color as 'black'
    self.contour_show_marker = True
    self.contour_marker_color = 'red'
    self.contour_smooth = 50
    self.contour_range_x = 0.3, 1.5
    self.contour_range_y = 0.5, 1.5



# Optimization
import numpy as np


def optimize(self):
    import copy
    import math
    from scipy import optimize  # type: ignore

    self.success = False

    if len(self.swrc) != 2:
        self.message = 'Error: No data of soil water retention curve.'
        return

    try:
        a = self.ini[0][0]
        gl = True
    except BaseException:
        gl = False
    if gl:
        f = self.multi_ini()
        if f.success:
            self.ini = f.fitted
        else:
            self.ini = f.ini

    self.mean_theta = np.average(self.swrc[1])
    self.var_theta = np.average((self.swrc[1] - self.mean_theta)**2)
    b = self.b_func()

    for c in sorted(self.const, reverse=True):
        b = b[:c[0] - 1] + b[c[0]:]

    if len(self.unsat) == 2:
        self.ht_only = False
        if min(self.unsat[1]) <= 0:
            self.message = 'Error: K should be positive.'
            return
        if 0 in self.f_hk(self.ini, self.unsat[0]):
            self.message = 'Overflow error: K(h) is too small and calculated as 0'
            return
        a = (self.swrc[0], self.swrc[1], self.unsat[0], self.unsat[1])
        cost = self.total_cost
        self.mean_k = np.average(self.unsat[1])
        self.var_k = np.average((self.unsat[1] - self.mean_k)**2)
        self.mean_ln_k = np.average(np.log(self.unsat[1]))
        self.var_ln_k = np.average(
            (np.log(self.unsat[1]) - self.mean_ln_k)**2)
    else:
        self.ht_only = True
        a = self.swrc
        cost = self.residual_ht
        for c in sorted(self.p_k_only, reverse=True):
            b = b[:c] + b[c + 1:]

    b = tuple(zip(*b))

    if self.debug:
        print('ini = {0}\nbounds = {1}'.format(
            self.ini, b))  # for debugging

    ini = self.ini
    success = False
    for ftol in self.lsq_ftol:
        result = optimize.least_squares(
            cost, ini, jac=self.lsq_jac, method=self.lsq_method, loss=self.lsq_loss,
            ftol=ftol, max_nfev=self.lsq_max_nfev, bounds=b, verbose=self.lsq_verbose, args=a)
        if result.success:
            ini = result.x
            success = True
            prev_result = copy.deepcopy(result)
        else:
            if success:
                result = copy.deepcopy(prev_result)
            break

    self.success = result.success  # True if convergence criteria is satisfied
    if not self.success:
        self.fitted = []
        self.message = result.message  # Verbal description of the termination reason
        return

    self.fitted = result.x  # Fitted parameters
    n = result.fun.size  # sample size
    k = self.fitted.size  # number of paramteres
    if self.ht_only:
        self.mse_ht = np.average(
            self.residual_ht(self.fitted, *self.swrc)**2)
        self.se_ht = math.sqrt(self.mse_ht)  # Standard error
        self.r2_ht = 1 - self.mse_ht / self.var_theta  # Coefficient of determination
        self.aic_ht = n * np.log(self.mse_ht) + 2 * k  # AIC
        if n-k-1 > 0:
            self.aicc_ht = self.aic_ht + 2*k*(k+1)/(n-k-1) # Corrected AIC
        self.message = self.format(
            self.param_ht, False).format(*self.fitted, self.r2_ht)
    else:
        p = list(self.fitted)
        for c in sorted(self.p_k_only, reverse=True):
            p = p[:c] + p[c + 1:]
        self.mse_ht = np.average(
            self.residual_ht(p, *self.swrc)**2)
        self.se_ht = math.sqrt(self.mse_ht)  # Standard error
        self.r2_ht = 1 - self.mse_ht / self.var_theta  # Coefficient of determination
        self.aic_ht = n * np.log(self.mse_ht) + 2 * k  # AIC
        if n-k-1 > 0:
            self.aicc_ht = self.aic_ht + 2*k*(k+1)/(n-k-1) # Corrected AIC
        self.mse_ln_hk = np.average(
            self.residual_ln_hk(self.fitted, *self.unsat)**2)
        self.se_ln_hk = math.sqrt(self.mse_ln_hk)  # Standard error
        self.r2_ln_hk = 1 - self.mse_ln_hk / self.var_ln_k  # Coefficient of determination
        self.aic_ln_hk = n * np.log(self.mse_ln_hk) + 2 * k  # AIC
        self.aicc_ln_h = self.aic_ln_hk + 2*k*(k+1)/(n-k-1) # Corrected AIC
        self.message = self.format(
            self.param, DualFitting=True).format(*self.fitted, self.r2_ht, self.r2_ln_hk)


def multi_ini(self):
    import copy
    import itertools
    comb = list(itertools.product(*self.ini))
    max_cost = -100000
    ftol = self.lsq_ftol
    self.lsq_ftol = self.lsq_ftol_global
    for ini in comb:
        self.ini = ini
        self.optimize()
        if self.success:
            if len(self.unsat) == 2:
                cost = self.r2_ht + self.r2_ln_hk
            else:
                cost = self.r2_ht
            if cost > max_cost:
                max_cost = cost
                max_f = copy.deepcopy(self)
    self.lsq_ftol = ftol
    if max_cost == -100000:
        return self
    return max_f


def format(self, param, ShowR2=True, DualFitting=False):
    format = ''
    count = 0
    for i in param:
        format += i + ' = {' + str(count) + ':' + \
            self.output_format[i] + '} '
        count += 1
    if ShowR2:
        if DualFitting:
            format += 'R2 q = {' + str(count) + \
                ':' + self.r2_format + \
                '} R2 logK = {' + str(count + 1) + ':' + self.r2_format + '}'
        else:
            format += 'R2 = {' + str(count) + ':' + self.r2_format + '}'
    return format


def residual_ht(self, p, x, y):
    return self.f_ht(p, x) - y


def residual_ln_hk(self, p, x, y):
    return np.log(self.f_hk(p, x) / y)


def residual_log10_hk(self, p, x, y):
    return np.log10(self.f_hk(p, x) / y)


def f_r2_ht(self, p, x, y):
    mse_ht = np.average(self.residual_ht(p, x, y)**2)
    return 1 - mse_ht / self.var_theta


def f_r2_ln_hk(self, p, x, y):
    mse_ln_hk = np.average(self.residual_ln_hk(p, x, y)**2)
    return 1 - mse_ln_hk / self.var_ln_k


def p_ht(self, p):
    p = list(p)
    for c in self.p_k_only:
        if c + 2 > len(p):
            p = p[:c]
        else:
            p = p[:c] + p[c + 1:]
    return p


def total_cost(self, p, x1, y1, x2, y2):
    r2 = self.f_r2_ht(self.p_ht(p), x1, y1) + self.f_r2_ln_hk(p, x2, y2)
    return 2 - r2



# dual-BC model
import numpy as np


def init_model_bc2(self):
    self.model['DB'] = self.model['dual-BC'] = self.model['bc2f'] = {
        'function': (self.bc2f, self.bc2f_k),
        'bound': self.bound_bc2f,
        'param': ['qs', 'qr', 'w1', 'hb1', 'l1', 'hb2', 'l2', 'Ks', 'p', 'q', 'r'],
        'k-only': [7, 8, 9, 10]
    }
    self.model['DBCH'] = self.model['DBC'] = self.model['dual-BC-CH'] = self.model['bc2'] = {
        'function': (self.bc2, self.bc2_k),
        'bound': self.bound_bc2,
        'get_init': self.get_init_bc2,
        'get_wrf': self.get_wrf_bc2,
        'param': ['qs', 'qr', 'hb', 'hc', 'l1', 'l2', 'Ks', 'p', 'q', 'r'],
        'k-only': [6, 7, 8, 9]
    }
    self.model['bc2ca'] = {
        'function': (self.bc2, self.bc2ca_k),
        'bound': self.bound_bc2,
        'get_init': self.get_init_bc2,
        'get_wrf': self.get_wrf_bc2,
        'param': ['qs', 'qr', 'hb', 'hc', 'l1', 'l2', 'Ks', 'p', 'a', 'r'],
        'k-only': [6, 7, 8, 9]
    }


def bound_bc2f(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_hb, self.b_lambda1, self.b_hb2,
            self.b_lambda2, self.b_ks, self.b_p, self.b_q, self.b_r]


def bc2f(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.bc2f_se(p, x) * (p[0] - p[1]) + p[1]


def bc2f_se(self, p, x):
    qs, qr, w1, hb1, l1, hb2, l2 = p
    s1 = np.where(x < hb1, 1, (x / hb1) ** (-l1))
    s2 = np.where(x < hb2, 1, (x / hb2) ** (-l2))
    return w1 * s1 + (1 - w1) * s2


def bc2f_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w1, hb1, l1, hb2, l2, ks, p, q, r = par
    w1b1 = w1 / hb1 ** q / (q / l1 + 1)
    w2b2 = (1 - w1) / hb2 ** q / (q / l2 + 1)
    s1 = np.where(x < hb1, 1, (x / hb1) ** (-l1 - q))
    s2 = np.where(x < hb2, 1, (x / hb2) ** (-l2 - q))
    bunshi = w1b1 * s1 + w2b2 * s2
    bunbo = w1b1 + w2b2
    return ks * self.bc2f_se(par[:7], x)**p * (bunshi / bunbo)**r

# dual-BC-CH model


def bound_bc2(self):
    return [self.b_qs, self.b_qr, self.b_hb, self.b_hc, self.b_lambda1,
            self.b_lambda2, self.b_ks, self.b_p, self.b_q, self.b_r]


def bc2(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.bc2_se(p, x) * (p[0] - p[1]) + p[1]


def bc2_se(self, p, x):
    w = 1 / (1 + (p[3] / p[2])**(p[5] - p[4]))
    # Ignore runtime warning, because divide by zero is warned when x=0
    import warnings
    warnings.simplefilter('ignore', category=RuntimeWarning)
    s1 = (x / p[2]) ** (-p[4])
    s2 = (x / p[2]) ** (-p[5])
    return np.where(x < p[2], 1, w * s1 + (1 - w) * s2)


def bc2_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, hb, hc, l1, l2, ks, p, q, r = par
    w1 = 1 / (1 + (hc / hb)**(l2 - l1))
    w1b1 = w1 / (q / l1 + 1)
    w2b2 = (1 - w1) / (q / l2 + 1)
    s1 = np.where(x < hb, 1, (x / hb) ** (-l1 - q))
    s2 = np.where(x < hb, 1, (x / hb) ** (-l2 - q))
    bunshi = w1b1 * s1 + w2b2 * s2
    bunbo = w1b1 + w2b2
    return ks * self.bc2_se(par[:6], x)**p * (bunshi / bunbo)**r


def get_init_bc2(self):  # hb, hc, l1, l2
    from .unsatfit import Fit
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    hb, l = f.get_init_bc()
    min_hb = min(min(x[np.nonzero(x)]) / 5, hb / 2)
    max_hb, min_hc, hc, max_hc = hb * (0.8, 0.7, 0.3, 0.1) ** (-1 / l)
    f.b_hb = (min_hb, max_hb)
    f.b_hc = (min_hc, max_hc)
    i = sum(x < max_hc)
    if len(x) - i > 1:
        w = 1 / (1 + (hc / hb)**(-l))
        x = np.log(x[i:] / hb)
        y = -np.log(y[i:] / (1 - w))
        l2 = self.linear_regress(x, y)
        if l2 < 0.01:
            l2 = 0.01
    else:
        l2 = 0
    f.set_model('bc2', const=[[1, 1], [2, 0], [3, hb], [4, hc], [6, l2]])
    f.ini = (l)
    f.optimize()
    if f.success:
        l1, = f.fitted
    else:
        l1, = l,
    f.set_model('bc2', const=[[1, 1], [2, 0]])
    f.ini = (hb, hc, l1, l2)
    f.b_lambda1 = (min(l, l1), l1 + 2)
    f.b_lambda2 = (0, max(l, l2))
    f.optimize()
    if f.success:
        return f.fitted
    f.ini = (hb, hb, l, l)
    f.b_lambda1 = (0, np.inf)
    f.b_lambda2 = (0, np.inf)
    f.optimize()
    if f.success:
        return f.fitted
    return f.ini


def get_wrf_bc2(self):
    from .unsatfit import Fit
    f = Fit()
    f.swrc = self.swrc
    f.debug = self.debug
    hb, hc, l1, l2 = f.get_init_bc2()
    f.set_model('bc2', const=['qr=0'])
    f.ini = (max(f.swrc[1]), hb, hc, l1, l2)
    f.optimize()
    if f.success:
        return (f.fitted[0], 0, *f.fitted[1:])
    return (f.ini[0], 0, *f.ini[1:])

# dual-BC-CH constant a model


def bc2ca_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, hb, hc, l1, l2, ks, p, a, r = par
    q = (a - p * l2) / r - l2
    w1 = 1 / (1 + (hc / hb)**(l2 - l1))
    w1b1 = w1 / (q / l1 + 1)
    w2b2 = (1 - w1) / (q / l2 + 1)
    s1 = np.where(x < hb, 1, (x / hb) ** (-l1 - q))
    s2 = np.where(x < hb, 1, (x / hb) ** (-l2 - q))
    bunshi = w1b1 * s1 + w2b2 * s2
    bunbo = w1b1 + w2b2
    return ks * self.bc2_se(par[:6], x)**p * (bunshi / bunbo)**r



# dual-KO model
import numpy as np


def init_model_ln2(self):
    self.model['DK'] = self.model['dual-KO'] = self.model['ln2'] = {
        'function': (self.ln2, self.ln2_k),
        'bound': self.bound_ln2,
        'param': ['qs', 'qr', 'w1', 'hm1', 'sigma1', 'hm2', 'sigma2', 'Ks', 'p', 'q', 'r'],
        'k-only': [7, 8, 9, 10]
    }
    self.model['DKCH'] = self.model['dual-KO-CH'] = self.model['ln2ch'] = {
        'function': (self.ln2ch, self.ln2ch_k),
        'bound': self.bound_ln2ch,
        'param': ['qs', 'qr', 'w1', 'hm1', 'sigma1', 'sigma2', 'Ks', 'p', 'q', 'r'],
        'k-only': [6, 7, 8, 9]
    }


def bound_ln2(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_hm1, self.b_sigma,
            self.b_hm2, self.b_sigma, self.b_ks, self.b_p, self.b_q, self.b_r]


def ln2(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.ln2_se(p, x) * (p[0] - p[1]) + p[1]


def ln2_se(self, p, x):
    s1 = self.ln_se([p[3], p[4]], x)
    s2 = self.ln_se([p[5], p[6]], x)
    return p[2] * s1 + (1 - p[2]) * s2


def ln2_k(self, p, x):
    from scipy.stats import norm
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, hm1, s1, hm2, s2, ks, p, q, r = par
    w1b1 = w * np.exp((q * s1)**2 / 2) / (hm1 ** q)
    w2b2 = (1 - w) * np.exp((q * s2)**2 / 2) / (hm2 ** q)
    q1 = 1 - norm.cdf(np.log(x / hm1) / s1 + q * s1)
    q2 = 1 - norm.cdf(np.log(x / hm2) / s2 + q * s2)
    bunshi = w1b1 * q1 + w2b2 * q2
    bunbo = w1b1 + w2b2
    return ks * self.ln2_se(par[:7], x)**p * (bunshi / bunbo)**r

# dual-KO-CH model


def bound_ln2ch(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_hm1, self.b_sigma,
            self.b_sigma, self.b_ks, self.b_p, self.b_q, self.b_r]


def ln2ch(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.ln2ch_se(p, x) * (p[0] - p[1]) + p[1]


def ln2ch_se(self, p, x):
    s1 = self.ln_se([p[3], p[4]], x)
    s2 = self.ln_se([p[3], p[5]], x)
    return p[2] * s1 + (1 - p[2]) * s2


def ln2ch_k(self, p, x):
    from scipy.stats import norm
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, hm1, s1, s2, ks, p, q, r = par
    w1b1 = w * np.exp((q * s1)**2 / 2) / (hm1 ** q)
    w2b2 = (1 - w) * np.exp((q * s2)**2 / 2) / (hm1 ** q)
    q1 = 1 - norm.cdf(np.log(x / hm1) / s1 + q * s1)
    q2 = 1 - norm.cdf(np.log(x / hm1) / s2 + q * s2)
    bunshi = w1b1 * q1 + w2b2 * q2
    bunbo = w1b1 + w2b2
    return ks * self.ln2ch_se(par[:6], x)**p * (bunshi / bunbo)**r
